# This file is part of librefdatool. librefdatool is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2013 Javier D. Garcia-Lasheras
#

import sys
import os
import subprocess
import csv

import numpy as np
from scipy import signal

from bitstring import BitArray

from snippets import *


class Simcore:
    '''This is a temporary class
    Should it be removed as the development advances??
    '''

    def simfilter(self, b, a, x, 
            structure, model,
            busX, busY, busC, 
            scalingX, scalingC):

        self.structure = structure
        self.model = model

        self.busX = busX
        self.busY = busY
        self.busC = busC

        self.scalingX = scalingX
        self.scalingC = scalingC
        
        # Define order values
        self.M = len(b)
        self.N = len(a)


        # Convert coef to fixed point

        self.coefB = b
        self.icoefB = self._quantizer(b, self.busC)
        
        # The A coefficients calculated with scipy uses a different convention and sign changes!!!!    
        a = -1 * a
        
        self.coefA = a
        self.icoefA = self._quantizer(a, self.busC)
        
        print('Coef B')
        print(self.icoefB)
        print('Coef A')
        print(self.icoefA)
        

        # Set DUT
        self._generateDUT()
        self._generateTB(x)

        return self._simulation()


    #***********************************************#
    #    Generate DUT
    #***********************************************#


    def _generateDUT(self):


            # All the products has an extra integer (the 2^1 factor)
            # A fixed point number is comprised by integer (P) and fractional bits (Q)
            # When you calculate the C product of A and B, then => PC = PA + PB; QC = QA + QB;
            # We are dealing with P=1 for all our numbers, so we can discard the resulting MSB.
            # By this way, the product of A and B limited to [-1,1) has a width of WC = WA + WB - 1
            
            # NOTE: we need to consider the possibility of P>1 !! (there are bigger than 1 coefficients) 
            # We are going to allow the full fixed point specification for X, Y and C!!
            
            # In the Sum S = A + B, the worst case for P is 1 + max(PA, PB).
            # For QS, we have the must of QA being equal to QB, so QA = QB = QS
            # As we have PA = PB = 1, PS would be 1 if there is not overflow. 
            # In any case, we can discard the overflow and assume that WC = WA = WS
            
            # In order to detect the overflow in the sum S = A + B, WA = WB:
            # If MSB(A) != MSB(B) => overflow cannot exist
            # if MSB(A)=MSB(B)=1 && MSB(S)=0 => overflow in the minimum range
            # if MSB(A)=MSB(B)=0 && MSB(S)=1 => overflow in the maximum range

        busX = self.busX[0] + self.busX[1]
        busY = self.busY[0] + self.busY[1]
        busC = self.busC[0] + self.busC[1]
        
        M = self.M
        N = self.N
        
        hdl = snippets()
        
        if self.model == 'vhdl':  

            structure = self.structure
            name = self.structure
            realization='direct_form_1_transposed'

            # Code generation for a Transposed Direct Form FIR
            hdlFile = open('{!s}.vhd'.format(name),'w')

            hdlFile.write(hdl.libraries())
            hdlFile.write('\n')
            
            hdlFile.write(hdl.entity(name, busX, busY))
            hdlFile.write('\n')
            
            hdlFile.write(hdl.architectureHeader(name, realization))
            hdlFile.write('\n')
            
            
            # *** DECLARE INTERNAL SIGNALS ***
            
            if realization == 'direct_form_1_transposed':
            
                hdlFile.write("-- Z^-1 delay blocks\n")
                # Z blocks in the b coefficients side
                for ii in range(self.M-1):
                    hdlFile.write('signal zb{!s}, zb{!s}_next: signed({!s} downto 0);\n'.format(ii, ii, busX + busC -1 -1))
                hdlFile.write("\n")
                
                # Z blocks in the a coefficients side
                for ii in range(self.N-1):
                    if ii != 0:
                        hdlFile.write('signal za{!s}, za{!s}_next: signed({!s} downto 0);\n'.format(ii, ii, busX + busC -1 -1))
                    else:
                        hdlFile.write('signal za{!s}, za{!s}_next: signed({!s} downto 0);\n'.format(ii, ii, busX -1))
                hdlFile.write("\n")
            
            
            hdlFile.write("-- Filter constants\n")
            for ii in range(self.M):               
                hdlFile.write('signal b{!s}: signed({!s} downto 0); \n'.format(ii, busC -1))
            hdlFile.write("\n")
            for ii in range(self.N):   
                if ii != 0:            
                    hdlFile.write('signal a{!s}: signed({!s} downto 0); \n'.format(ii, busC -1))
            hdlFile.write("\n")

            hdlFile.write("-- Filter Adders\n")
            for ii in range(self.M-1):
                hdlFile.write('signal sb{!s}: signed({!s} downto 0);\n'.format(ii, busX + busC -1 -1))
            hdlFile.write("\n")
            for ii in range(self.N-1):
                if ii != 0:  
                    hdlFile.write('signal sa{!s}: signed({!s} downto 0);\n'.format(ii, busX + busC -1 -1))
                else:  
                    hdlFile.write('signal sa{!s}: signed({!s} downto 0);\n'.format(ii, busX -1))
            hdlFile.write("\n")

            hdlFile.write("-- Filter Products\n")
            for ii in range(self.M):
                hdlFile.write('signal pb{!s}: signed({!s} downto 0);\n'.format(ii, busX + busC -1 -1))
            hdlFile.write("\n")    
            for ii in range(self.N):
                if ii != 0:
                    hdlFile.write('signal pa{!s}: signed({!s} downto 0);\n'.format(ii, busX + busC -1 -1))
            hdlFile.write("\n")  
            
            # Intermediate signal for df1: assume that the width is the same that in x!
            # NOTE: we have set v as equal to X, but this may lead to overflow troubles!!!!!!
            hdlFile.write('signal v: signed({!s} downto 0);\n\n'.format(busX -1))
               
        
            hdlFile.write("-- Begin Architecture\n")
            hdlFile.write("begin\n\n")

            hdlFile.write(hdl.sequentialBlock(M, N, clkEdge='posedge', rstActive='high'))
            hdlFile.write("\n")

            hdlFile.write('-- Arithmetics\n')
            hdlFile.write("\n")

        
            # Coefficients
            for ii in range(self.M):
                hdlFile.write('b{!s} <= \"{!s}\"; -- {!s}\n'.format(ii, self.icoefB[ii], self.coefB[ii]))
            hdlFile.write("\n")
            for ii in range(self.N):
                if ii != 0:
                    hdlFile.write('a{!s} <= \"{!s}\"; -- {!s}\n'.format(ii, self.icoefA[ii], self.coefA[ii]))
            hdlFile.write("\n")  


            # Products
            for ii in range(self.M):
                hdlFile.write('pb{!s} <= v * b{!s};\n'.format(ii, ii))
            hdlFile.write("\n")      
            for ii in range(self.N):
                if ii != 0:
                    hdlFile.write('pa{!s} <= v * a{!s};\n'.format(ii, ii))
            hdlFile.write("\n")              

       
            # Sums
            for ii in range(self.M-1):
                hdlFile.write('sb{!s} <= pb{!s} + zb{!s};\n'.format(ii, ii, ii))
            hdlFile.write("\n\n")  
            for ii in range(self.N-1):
                if ii == 0:
                    hdlFile.write('sa{!s} <= sig_in + za{!s};\n'.format(ii, ii))
                else:
                    hdlFile.write('sa{!s} <= pa{!s} + za{!s};\n'.format(ii, ii, ii))
            hdlFile.write("\n\n")  


            hdlFile.write('-- Signal link, trunk, overflow, fixed point...\n')
            for ii in range(self.M):
                if ii == 0:
                    # Why always is a Zero as MSB in si? --> this is because the input is always a positive number?????
                    if self.M == 1:
                        hdlFile.write(hdl.convertFixedPoint('pb{!s}'.format(ii), self.busC[0] + self.busX[0] -1, self.busC[1] + self.busX[1], 'sig_out', self.busY[0], self.busY[1]))
                    else:
                        hdlFile.write(hdl.convertFixedPoint('sb{!s}'.format(ii), self.busC[0] + self.busX[0] -1, self.busC[1] + self.busX[1], 'sig_out', self.busY[0], self.busY[1]))
                elif ii == (self.M-1):
                    hdlFile.write('zb{!s}_next <= pb{!s};\n'.format(ii-1, ii))
                else:
                    hdlFile.write('zb{!s}_next <= sb{!s};\n'.format(ii-1, ii))

            for ii in range(self.N):
                if ii == 0:
                    # Why always is a Zero as MSB in si? --> this is because the input is always a positive number?????
                    if self.N == 1:
                        hdlFile.write('v <= sig_in;\n')
                    else:
                        hdlFile.write('v <= sa0;\n') 
                elif ii == 1:
                    if ii != (self.N-1):
                        hdlFile.write(hdl.convertFixedPoint('sa{!s}'.format(ii), self.busC[0] + self.busX[0] -1, self.busC[1] + self.busX[1], 'za{!s}_next'.format(ii-1), self.busX[0], self.busX[1]))  
                    else:
                        hdlFile.write(hdl.convertFixedPoint('pa{!s}'.format(ii), self.busC[0] + self.busX[0] -1, self.busC[1] + self.busX[1], 'za{!s}_next'.format(ii-1), self.busX[0], self.busX[1]))
                elif ii == (self.N-1):
                    hdlFile.write('za{!s}_next <= pa{!s};\n'.format(ii-1, ii))
                else:
                    hdlFile.write('za{!s}_next <= sa{!s};\n'.format(ii-1, ii))        

            hdlFile.write("\n\n") 

            hdlFile.write('end {!s};\n'.format(realization))
           
            hdlFile.close()
            


        elif self.model == 'verilog':  

            structure = self.structure
            realization='direct_form_transposed'

            # Verilog Code generation for a Transposed Direct Form FIR
            hdlFile = open('{!s}.v'.format(structure),'w')

            hdlFile.write("// This is a transposed direct form test\n\n")
            
            hdlFile.write("`timescale 1ns/10ps\n\n")
            
            hdlFile.write("module {!s} (\n".format(structure))
            hdlFile.write("    clk,\n")
            hdlFile.write("    rst,\n")
            hdlFile.write("    sig_in,\n")
            hdlFile.write("    sig_out\n")
            hdlFile.write(");\n\n")

            
            hdlFile.write('// Architecture depends on selected structure\n')
            hdlFile.write('// It would be nice to have target optimizations too in the future\n')
            hdlFile.write('//\n')
            hdlFile.write('// example, transposed direct form for FIR:\n')
            hdlFile.write('//\n')
            hdlFile.write('//              z^-1  z^-1             z^-1  z^-1\n')
            hdlFile.write('//            o-->--o-->--o-->- - - --o-->--o-->--o-->--o y[n]\n')
            hdlFile.write('//            |     |     |           |     |     | \n')
            hdlFile.write('//            ^Bm   ^Bm-1 ^Bm-2       ^B2   ^B1   ^B0\n')
            hdlFile.write('//            |     |     |           |     |     |\n')
            hdlFile.write('// x[n] o-->--o-->--o-->--o-->- - - --o-->--o-->--\n')
            hdlFile.write('//\n\n')
            
            hdlFile.write("input clk;\n")
            hdlFile.write("input rst;\n")
            hdlFile.write("input signed [{!s}:0] sig_in;\n".format(self.busX-1))
            hdlFile.write("output signed [{!s}:0] sig_out;\n".format(self.busY-1))
            
            hdlFile.write("// Z^-1 delay blocks\n")
            hdlFile.write('reg signed [{!s}:0] z [0:{!s}];\n'.format(self.busX + self.busC -1, len(self.icoef)-1))
            hdlFile.write('wire signed [{!s}:0] z_next [0:{!s}];\n'.format(self.busX + self.busC -1, len(self.icoef)-1))
            hdlFile.write("\n")
            
            hdlFile.write("// Filter constants\n")
            for ii in range(self.M):               
                hdlFile.write('wire signed [{!s}:0] b{!s};\n'.format(self.busC -1, ii))
            hdlFile.write("\n")

            hdlFile.write("// Filter Adders\n")
            for ii in range(self.M-1):
                hdlFile.write('wire signed [{!s}:0] s{!s};\n'.format(self.busX + self.busC -1, ii))
            hdlFile.write("\n")

            hdlFile.write("// Filter Products\n")
            for ii in range(self.M):
                hdlFile.write('wire signed [{!s}:0] p{!s};\n'.format(self.busX + self.busC -1, ii))
            hdlFile.write("\n")  
            
            
            hdlFile.write("always @(posedge clk) begin: {!s}\n".format(realization))
            hdlFile.write("    if (rst == 1'b1) begin\n")
            for ii in range(self.M-1):
                hdlFile.write('        z[{!s}] <= 0;\n'.format(ii))
            hdlFile.write("    end\n")
            hdlFile.write("    else begin\n")
            for ii in range(self.M-1):
                hdlFile.write('        z[{!s}] <= z_next[{!s}];\n'.format(ii, ii))
            hdlFile.write("    end\n")
            hdlFile.write("end\n")
            hdlFile.write("\n")
            
            hdlFile.write('// Arithmetics\n')
            hdlFile.write("\n")
            for ii in range(self.M):
                hdlFile.write('assign b{!s} = {!s}\'b{!s}; // {!s}\n'.format(ii, self.busC, self.icoef[ii], self.coef[ii]))
            hdlFile.write("\n")  
            for ii in range(self.M):
                hdlFile.write('assign p{!s} = sig_in * b{!s};\n'.format(ii, ii))
            hdlFile.write("\n")  
            for ii in range(self.M-1):
                hdlFile.write('assign s{!s} = p{!s} + z[{!s}];\n'.format(ii, ii, ii))
            hdlFile.write("\n\n")  

            hdlFile.write('// Signal link, trunk, overflow...\n')
            for ii in range(self.M):
                if ii == 0:
                    # Why always is a Zero as MSB in si? --> this is because the input is always a positive number?????
                    hdlFile.write('assign sig_out = s{!s}[{!s}:{!s}];\n'.format(ii, self.busX + self.busC -1, self.busX + self.busC - self.busY))
                elif ii == (len(self.coef)-1):
                    hdlFile.write('assign z_next[{!s}] = p{!s};\n'.format(ii-1, ii))
                else:
                    hdlFile.write('assign z_next[{!s}] = s{!s};\n'.format(ii-1, ii))
            hdlFile.write("\n\n")  
            
            hdlFile.write("endmodule\n")
            hdlFile.write("\n")
           
            hdlFile.close()

 
        else:
            print("Error: simulation mode not recognized\n");

    
        return 0


    def _generateTB(self, x):
        
            # TBD: the process of input signal and output value parse
            # must be aligned and configurable
            # This Verilog file is common for verilog and vhdl DUTs in IVerilog
            # If we want to add support for more simulators, then would be necessary VHDL

            # Code generation for the testbench - myhdl link 

            busX = self.busX[0] + self.busX[1]
            busY = self.busY[0] + self.busY[1]
            busC = self.busC[0] + self.busC[1]
            
            structure = self.structure
            
            tbFile = open('tb_{!s}.v'.format(structure),'w')

            tbFile.write("module tb_{!s};\n".format(structure))
            tbFile.write("\n")
            tbFile.write("reg clk;\n")
            tbFile.write("reg rst;\n")
            tbFile.write("reg [{!s}:0] sig_in;\n".format(busX-1))
            tbFile.write("wire [{!s}:0] sig_out;\n".format(busY-1))
            tbFile.write("\n")
            tbFile.write("{!s} dut(\n".format(structure))
            tbFile.write("    clk,\n")
            tbFile.write("    rst,\n")
            tbFile.write("    sig_in,\n")
            tbFile.write("    sig_out\n")
            tbFile.write(");\n")
            tbFile.write("\n")
            
            clockPosWidth = 25
            clockNegWidth = 25
            tbFile.write("always begin\n")
            tbFile.write("    clk <= 1;\n")
            tbFile.write("    #{!s};\n".format(clockNegWidth))
            tbFile.write("    clk <= 0;\n")
            tbFile.write("    $display(\"%d, %b, %b\", $time, sig_in, sig_out);\n")
            tbFile.write("    #{!s};\n".format(clockPosWidth))
            tbFile.write("end\n")
            tbFile.write("\n")
            tbFile.write("\n")
            tbFile.write("\n")
            tbFile.write("\n")            
            
           
            tbFile.write("initial begin\n")
            tbFile.write("    $dumpfile(\"{!s}.vcd\");\n".format(structure))
            tbFile.write("    $dumpvars;\n")
            tbFile.write("end\n")
            tbFile.write("\n")
            
            
            tbFile.write("initial begin\n")
            tbFile.write('    sig_in <= {!s}\'b{!s}; // signal is initially zero\n'.format(busX, 0))
            tbFile.write("    rst <= 1'b1; #100;\n")
            tbFile.write("    rst <= 1'b0; #150;\n")
                   
                   
            xb = self._quantizer(x, self.busX, staircase='midtread')   
            for ii in range(len(x)):
                tbFile.write('    #50; sig_in <= {!s}\'b{!s}; // {!s}\n'.format(busX, xb[ii], x[ii]))
            tbFile.write("    $finish;\n")
            tbFile.write("end\n")
            tbFile.write("\n")
            
            
            tbFile.write("\n")
            tbFile.write("endmodule\n\n")   
            
            tbFile.close()


    def _quantizer(self, signal, bus, xm=1, staircase= 'midtread'):
        
        # Normalization to the [-1, 1) input range 
        signalNorm = signal/xm 
        
        # Generate the fixed point code in two's complement
        # 
        # xb = -a0*2^0 + a1*2^(-1) + a2*2^(-2) + ... + aB*2^(-B)
        # bus = B + 1; 
        
        B = bus[1] + bus[0] - 1
        
        # Generate an 
        
        if staircase == 'midtread':
            delta = 1/2
        elif staircase == 'midriser':
            delta = 0
        else:
            print("Staircase model for quantizer unknown: using default")
            delta = 1/2
        
        # Convert the normalized signal to a integer value
        signalInt = np.floor(signalNorm*(2**(bus[1])) + delta)
        
        # Trunk if the integers cannot be represented using the bus width
        # Then assign the correspondent binary code
        xBinary = []
        for ii in range(len(signal)):
            if signalInt[ii] > (2**B -1):
                signalInt[ii] = 2**B - 1
            elif signalInt[ii] < (-1 * 2**B):
                signalInt[ii] = -1 * 2**B
            xBinary.append(np.binary_repr(int(signalInt[ii]), width=B+1))
                
        return xBinary
        
   
    def _simulation(self, simulator='iverilog'):

        structure = self.structure
        
        if self.model == 'verilog':
            dutExtension = 'v'
        elif self.model == 'vhdl':
            dutExtension = 'vhd'
        else:
            print('Unknown model: assuming verilog')
            dutExtension = 'v'        
        
        
        print('TESTPOINT_1')

        print('Cosimulation...')
        print('TEST iverilog -o {!s} {!s}.{!s} tb_{!s}.v'.format(structure, structure, dutExtension, structure))
        # Check if verilog/vhdl filter description is available TBD: make it a variable!!!!!!!!!!!!!!!!
        if not os.path.isfile('{!s}.vhd'.format(structure)):
            print('File not found: Run generation first')
            return None
        # Clean design
        if os.path.isfile('{!s}'.format(structure)):
            print('Existing compiled design: erase!! ')
            cmd = 'rm {!s}'.format(structure)
            os.system(cmd)
    
            
        # Running Icarus verilog

        # ****** TBD: Set the dut file extension according with the HDL language  
        cmd = 'iverilog -g2012 -o {!s} {!s}.{!s} tb_{!s}.v'.format(structure, structure, dutExtension, structure)
        os.system(cmd)

        cmd = 'vvp {!s}'.format(structure)

        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        (out, err) = proc.communicate()
            
        reportFile = open('{!s}_report.txt'.format(structure),'wb')
        reportFile.write(out) 
        reportFile.close()
            
        #print("program output: {!s}".format(out))
        print('TESTPOINT_4')
        
        parsedOutput = []
        
        with open('{!s}_report.txt'.format(structure),'r') as csvfile:
            csvString = csv.reader(csvfile, delimiter=',', quotechar='|', skipinitialspace=True)
            ii = 0
            for row in csvString:
                # Only the lines with three fields qualifies
                if len(row) == 3:
                    # Only the values beyond a time are considered valids samples
                    # Note that init and reset values must be neglected 
                    if int(row[0]) > 250:
                        a = BitArray(bin=row[2])
                        #print(a.int)
                        parsedOutput.append(a.int/float(2**(self.busY[1])  ))
                        ii = ii + 1
        
        # The fixed point format for Y is calculated in an automatic fashion!!!
        
        connect_sample_out = np.zeros(len(parsedOutput))
        for ii in range(len(parsedOutput)):
            connect_sample_out[ii] = parsedOutput[ii]
            
        return connect_sample_out
    
    

