# This file is part of librefdatool. librefdatool is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright (C) 2013 Javier D. Garcia-Lasheras
#


import sys
import os
import subprocess
import csv

import numpy as np
from scipy import signal

from bitstring import BitArray


class snippets:

# we are going to use dictionary-based string formatting

    def libraries(self):
        
        libString = (
            'library ieee;\n'
            'use ieee.std_logic_1164.all;\n'  
            'use ieee.numeric_std.all;\n\n'           
        ) 
        
        return libString       


    def entity(self, name, busX, busY):
        
        entityDict = {
            'Name': name,
            'MSBitX': busX - 1,
            'MSBitY': busY - 1
        }
        
        entityString = (
            'entity %(Name)s is\n'
            '    port (\n'
            '        clk : in std_logic;\n'
            '        rst : in std_logic;\n'
            '        sig_in : in signed(%(MSBitX)u downto 0);\n'
            '        sig_out : out signed(%(MSBitY)u downto 0)\n'
            '    );\n'
            'end entity;\n'             
        ) 
        
        return entityString % entityDict


    def architectureHeader(self, name, realization):
 
        #TBD: include the appropriated realization diagram

        archHeadDict = {
            'Name': name,
            'Realization': realization
        }        
        
        archHeadString = (
            'architecture %(Realization)s of %(Name)s is\n\n'
            '-- Architecture depends on selected realization\n'
            '-- TODO: this header should each different realization\n'
            '--\n'
            '-- example, transposed direct form for FIR:\n'
            '--\n'
            '--              z^-1  z^-1             z^-1  z^-1\n'
            '--            o-->--o-->--o-->- - - --o-->--o-->--o-->--o y[n]\n'
            '--            |     |     |           |     |     | \n'
            '--            ^Bm   ^Bm-1 ^Bm-2       ^B2   ^B1   ^B0\n'
            '--            |     |     |           |     |     |\n'
            '-- x[n] o-->--o-->--o-->--o-->- - - --o-->--o-->--\n'
            '--\n'                                                   
        ) 
 
        return archHeadString % archHeadDict
        


    def convertFixedPoint(self, nameX, pX, qX, nameY, pY, qY, overflow=False, language='vhdl'):

        # TBD: overflow handling
        # The syntax is not very handy when pX, qX, pY, qY, qDiff or pDiff are equal to 1 => (i downto i)???
       
        pDiff = pY - pX
        qDiff = qY - qX

        # INTEGER PART
        if pDiff == 0:
            # Destination integer part is == than the input one
            signExt = ''
            MSBpY = pY + qY - 1 
            LSBpY = qY
            MSBpX = pX + qX - 1
            LSBpX = qX
        elif pDiff > 0:
            signExt = '%(nameY)s({!s} downto {!s}) <= (others => %(nameX)s({!s}));\n'.format(pY+qY-1, pY+qY-pDiff, pX+qX-1)
            MSBpY = pY + qY - 1 - pDiff 
            LSBpY = qY
            MSBpX = pX + qX - 1
            LSBpX = qX
            # Destination integer part is > than the input one
            # sign extension in the extra Y bits!!
            # Now the rest of the significant bits
        else: # pdiff < 0
            signExt = ''
            MSBpY = pY + qY - 1 
            LSBpY = qY
            MSBpX = pX + qX - 1 + pDiff
            LSBpX = qX
            # Destination integer part is < than the input one
            # We should be carefull with overflow: TBD
            
        # FRACTIONAL PART (condition: qY and qX must be greater than zero???)
        if qDiff == 0:
            trunkExt = ''
            MSBqY = qY - 1 
            LSBqY = 0
            MSBqX = qX - 1
            LSBqX = 0
            # Destination fractional part is == than the input one    
        elif qDiff > 0:
            # Destination fractional part is > than the input one 
            # Fill with zeros the unused bits  
            trunkExt = '%(nameY)s({!s} downto 0) <= (others => \'0\');\n'.format(qDiff-1)
            MSBqY = qY - 1 
            LSBqY = qDiff
            MSBqX = qX - 1
            LSBqX = 0
        else: #qDiff < 0
            # Destination fractional part is < than the input one 
            trunkExt = ''
            MSBqY = qY - 1 
            LSBqY = 0
            MSBqX = qX - 1
            LSBqX = -1 * qDiff              

        convertFPDict = {
            'nameX': nameX,
            'nameY': nameY,
            'signExt': signExt,
            'MSBpY': MSBpY,
            'LSBpY': LSBpY,
            'MSBpX': MSBpX,
            'LSBpX': LSBpX,
            'trunkExt': trunkExt,
            'MSBqY': MSBqY,
            'LSBqY': LSBqY,
            'MSBqX': MSBqX,
            'LSBqX': LSBqX
        }
        
        convertFPString = (
            '%(signExt)s%(nameY)s(%(MSBpY)u downto %(LSBpY)u)'
            ' <= %(nameX)s(%(MSBpX)u downto %(LSBpX)u);\n'
            '%(trunkExt)s%(nameY)s(%(MSBqY)u downto %(LSBqY)u)'
            ' <= %(nameX)s(%(MSBqX)u downto %(LSBqX)u);\n'                   
        )            
                   
        return convertFPString % convertFPDict


    

    def sequentialBlock(self, M, N, clkEdge='posedge', rstActive='high'):
        # TBD: different rst and clk polarity
        # Now, they are true and posedge respectively
        
        # Check the arguments
        
        if clkEdge == 'posedge':
            boolEdge = '1'
        elif clkEdge == 'negedge':
            boolEdge = '0'
        else:
            sys.exit('ERROR: active clock edge not recognized')
            
        if rstActive == 'high':
            boolRst = '1'
        elif rstActive == 'low':
            boolRst = '0'
        else:
            sys.exit('ERROR: reset active level not recognized')


        seqDict = {
            'boolEdge': boolEdge,
            'boolRst': boolRst
        }                 

        if M >= 1 or N >= 1:
            
            seqString = (
                '-- Sequential block\n'
                'z_block: process (clk)\n'
                '    begin\n'
                '        if (clk\'event and clk = \'%(boolEdge)s\') then\n'
                '            if (rst = \'%(boolRst)s\') then\n'
            )
            for ii in range(M-1):
                seqString += '                zb%(ii)u <= 0;\n' % {'ii': ii}
            for ii in range(N-1):
                seqString += '                za%(ii)u <= 0;\n' % {'ii': ii}
            seqString += '            else\n'   
            for ii in range(M-1):
                seqString += '                zb%(ii)u <= zb%(ii)u_next;\n' % {'ii': ii}
            for ii in range(N-1):
                seqString += '                za%(ii)u <= za%(ii)u_next;\n' % {'ii': ii}
            seqString += (
                '            end if;\n' 
                '        end if;\n' 
                'end process;\n' 
            )            

        return seqString % seqDict



